

#ifndef _THUNK64_H_
#define _THUNK64_H_


typedef void * __ptr64 P64VOID;
typedef void * __ptr64 LP64VOID;
typedef const void * __ptr64 P64CVOID;
typedef const void * __ptr64 LP64CVOID;
typedef void * __ptr64 HANDLE64;
typedef void * __ptr64 HMODULE64;
typedef void * __ptr64 HINSTANCE64;
typedef char * __ptr64 P64STR;
typedef char * __ptr64 LP64STR;
typedef short * __ptr64 P64WSTR;
typedef short * __ptr64 LP64WSTR;

typedef NT_TIB64 * PTEB64_;
typedef NT_TIB64 * LPTEB64_;
typedef NT_TIB64 * __ptr64 P64TEB64_;
typedef NT_TIB64 * __ptr64 LP64TEB64_;

typedef struct _LIST_ENTRY64_ {
   struct _LIST_ENTRY64_ * __ptr64 Flink;
   struct _LIST_ENTRY64_ * __ptr64 Blink;
} LIST_ENTRY64_, *PLIST_ENTRY64_, *LPLIST_ENTRY64_, * __ptr64 P64LIST_ENTRY64_, * __ptr64 LP64LIST_ENTRY64_;

typedef struct _UNICODE_STRING64_ {
  USHORT Length;
  USHORT MaximumLength;
  DWORD align1;
  P64WSTR Buffer;
} UNICODE_STRING64_, *PUNICODE_STRING64_, *LPUNICODE_STRING64_, * __ptr64 P64UNICODE_STRING64_, * __ptr64 LP64UNICODE_STRING64_;

typedef struct _LDR_DATA_TABLE_ENTRY64_ {
  LIST_ENTRY64_ InLoadOrderLinks;
  LIST_ENTRY64_ InMemoryOrderLinks;
  LIST_ENTRY64_ InInitializationOrderLinks;
  P64VOID DllBase;
  P64VOID EntryPoint;
  ULONGLONG SizeOfImage;
  UNICODE_STRING64_ FullDllName;
  UNICODE_STRING64_ BaseDllName;
  ULONG Flags;
  USHORT LoadCount;
  USHORT TlsIndex;
  union {
    LIST_ENTRY64_ HashLinks;
    struct {
      P64VOID SectionPointer;
      ULONG CheckSum;
    } s;
  } u;
  ULONG TimeDateStamp;
  //etc...
} LDR_DATA_TABLE_ENTRY64_, *PLDR_DATA_TABLE_ENTRY64_, *LPLDR_DATA_TABLE_ENTRY64_, * __ptr64 P64LDR_DATA_TABLE_ENTRY64_, * __ptr64 LP64LDR_DATA_TABLE_ENTRY64_;

typedef struct _PEB_LDR_DATA64_ {
  DWORD Length;
  BYTE Initialized;
  BYTE Dummy[3];
  HANDLE64 SsHandle;
  LIST_ENTRY64_ InLoadOrderModuleList;
  LIST_ENTRY64_ InMemoryOrderModuleList;
  LIST_ENTRY64_ InInitializationOrderModuleList;
} PEB_LDR_DATA64_, *PPEB_LDR_DATA64_, *LPPEB_LDR_DATA64_, * __ptr64 P64PEB_LDR_DATA64_, * __ptr64 LP64PEB_LDR_DATA64_;

typedef struct _PEB64_ {
  BYTE InheritedAddressSpace;
  BYTE ReadImageFileExecOptions;
  BYTE BeingDebugged;
  BYTE SpareBool;
  HANDLE64 Mutant;
  HMODULE64 ImageBaseAddress;
  P64PEB_LDR_DATA64_ Ldr;
  //etc...
} PEB64_, *PPEB64_, *LPPEB64_, * __ptr64 P64PEB64_, * __ptr64 LP64PEB64_;



#if !defined(NTSTATUS)
#define NTSTATUS LONG
//typedef LONG NTSTATUS;
#endif


#define PAR64(x) ((ULONGLONG)(x))

//32bit pointers are sufficient in Wow64 (unable to dereference Ptr64)
/*
P64TEB64_ GetCurrentTEB64(VOID) {
  __asm {
    mov eax, gs:[30h]
    mov edx, gs:[34h]
  }
}

P64PEB64_ GetCurrentPEB64(VOID) {
  __asm {
    mov eax, gs:[60h]
    mov edx, gs:[64h]
  }
}
*/

PTEB64_ GetCurrentTEB64(VOID) {
  __asm {
    mov eax, gs:[30h]
  }
}

PPEB64_ GetCurrentPEB64(VOID) {
  __asm {
    mov eax, gs:[60h]
  }
}


LP64VOID __declspec(naked) __cdecl SysCall64(ULONGLONG dwServiceNumber, ULONG nParameters,/* ULONGLONG Parameter1,*/ ...) {
  dwServiceNumber; nParameters;
  __asm {
    push  esi
    push  edi
    lea   esi, [esp+24]
    push  cs
    call  Go32
    push  ss
    pop   ss
    lea   esp, [esi-24]
    pop   edi
    pop   esi
    ret
   Go32:
    push  0x33
    call  Go64
    dec   eax
    mov   edi, esp
    dec   eax       ;dec ecx
    and   esp, ~15  ;mov esp, esi
    mov   ecx, dword ptr [esi-4]
   Push5:
    cmp   ecx, 4
    jbe   Push4
    sub   ecx, 1
    push  dword ptr [esi+8*ecx]
    jmp   Push5
   Push4:
    dec   eax
    sub   esp, 32
    jecxz Push0
    dec   cl
    je    Push1
    dec   cl
    je    Push2
    dec   cl
    je    Push3
    dec   esp
    mov   ecx, [esi+8*3]
   Push3:
    dec   esp
    mov   eax, [esi+8*2]
   Push2:
    dec   eax
    mov   edx, [esi+8*1]
   Push1:
    dec   eax
    mov   ecx, [esi+8*0]
   Push0:
    mov   eax, dword ptr [esi-12]
    dec   esp
    mov   edx, ecx
    ;syscall
    _emit 0xf
    _emit 0x5
    dec   eax
    mov   esp, edi
    dec   eax
    mov   edx, eax
    dec   eax
    shr   edx, 32
   Go64:
    retf
  }
}

LP64VOID __declspec(naked) __cdecl ApiCall64(LP64CVOID lp64Api, ULONG nParameters,/* ULONGLONG Parameter1,*/ ...) {
  lp64Api; nParameters;
  __asm {
    push  esi
    push  edi
    lea   esi, [esp+24]
    push  cs
    call  Go32
    push  ss
    pop   ss
    lea   esp, [esi-24]
    pop   edi
    pop   esi
    ret
   Go32:
    push  0x33
    call  Go64
    dec   eax
    mov   edi, esp
    dec   eax       ;dec ecx
    and   esp, ~15  ;mov esp, esi
    mov   ecx, dword ptr [esi-4]
   Push5:
    cmp   ecx, 4
    jbe   Push4
    sub   ecx, 1
    push  dword ptr [esi+8*ecx]
    jmp   Push5
   Push4:
    dec   eax
    sub   esp, 32
    jecxz Push0
    dec   cl
    je    Push1
    dec   cl
    je    Push2
    dec   cl
    je    Push3
    dec   esp
    mov   ecx, [esi+8*3]
   Push3:
    dec   esp
    mov   eax, [esi+8*2]
   Push2:
    dec   eax
    mov   edx, [esi+8*1]
   Push1:
    dec   eax
    mov   ecx, [esi+8*0]
   Push0:
    call  dword ptr [esi-12]
    dec   eax
    mov   edx, eax
    dec   eax
    shr   edx, 32
    dec   eax
    mov   esp, edi
   Go64:
    retf
  }
}

LPCVOID WINAPI thGetProcAddress64(HMODULE hModule, LPCSTR lpProcName) {
//  PIMAGE_NT_HEADERS64 pPE;
  PIMAGE_EXPORT_DIRECTORY pExport;
  DWORD iNm, nNm, *pNm;
  #define pPE ((PIMAGE_NT_HEADERS64)((ULONG_PTR)hModule +((PIMAGE_DOS_HEADER)hModule)->e_lfanew))
/*
  pPE = (PIMAGE_NT_HEADERS64)((ULONG_PTR)hModule +((PIMAGE_DOS_HEADER)lpFile)->e_lfanew);
  if((pPE->Signature != IMAGE_NT_SIGNATURE) || (pPE->FileHeader.Machine != IMAGE_FILE_MACHINE_AMD64) || (pPE->OptionalHeader.Magic != IMAGE_NT_OPTIONAL_HDR64_MAGIC)) {
    return(NULL);
  }
  if(pPE->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT].Size == 0) {
    return(NULL);
  }
*/
  pExport = (PIMAGE_EXPORT_DIRECTORY)((ULONG_PTR)hModule +pPE->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_EXPORT].VirtualAddress);
  #undef pPE
  nNm = pExport->NumberOfFunctions;
  pNm = (PDWORD)((ULONG_PTR)hModule +pExport->AddressOfNames);
  for(iNm = 0; iNm < nNm; iNm++) {
    if(STRCMPI(lpProcName, (LPSTR)((ULONG_PTR)hModule +pNm[iNm])) == 0) {
      return((LPCVOID)((ULONG_PTR)hModule +((PDWORD)((ULONG_PTR)hModule +pExport->AddressOfFunctions))[((PWORD)((ULONG_PTR)hModule +pExport->AddressOfNameOrdinals))[iNm]]));
    }
  }
  return(NULL);
}

#endif
